From e306266b65c37f7905127f7f45f207a6a781e4c7 Mon Sep 17 00:00:00 2001
From: Sebastian Pop <spop@amazon.com>
Date: Thu, 29 Jul 2021 17:26:53 +0000
Subject: [PATCH] [arm64] port sub_ps

         sub_ps[  4x4]	10.92x 	 7.31     	 79.84
         sub_ps[  8x8]	25.58x 	 12.40    	 317.33
 [i420]  sub_ps[  4x4]	11.04x 	 7.60     	 83.94
 [i422]  sub_ps[  4x8]	12.91x 	 11.97    	 154.50
         sub_ps[16x16]	47.75x 	 27.14    	 1295.89
 [i420]  sub_ps[  8x8]	28.23x 	 11.94    	 337.09
 [i422]  sub_ps[ 8x16]	27.13x 	 24.28    	 658.73
         sub_ps[32x32]	54.61x 	 81.43    	 4446.65
 [i420]  sub_ps[16x16]	43.31x 	 26.63    	 1153.34
 [i422]  sub_ps[16x32]	49.62x 	 50.06    	 2484.23
         sub_ps[64x64]	55.10x 	 316.59   	 17444.26
 [i420]  sub_ps[32x32]	55.64x 	 81.55    	 4537.30
 [i422]  sub_ps[32x64]	55.71x 	 159.67   	 8894.43
---
 source/common/aarch64/asm-primitives.cpp |  17 +++
 source/common/aarch64/pixel-util.S       | 170 +++++++++++++++++++++++
 source/common/aarch64/pixel.h            |  40 ++++++
 3 files changed, 227 insertions(+)

diff --git a/source/common/aarch64/asm-primitives.cpp b/source/common/aarch64/asm-primitives.cpp
index c9b282a..5e57f31 100644
--- a/source/common/aarch64/asm-primitives.cpp
+++ b/source/common/aarch64/asm-primitives.cpp
@@ -413,6 +413,23 @@ void setupAssemblyPrimitives(EncoderPrimitives &p, int cpuMask)
         p.cu[BLOCK_16x16].cpy1Dto2D_shr = PFX(cpy1Dto2D_shr_16x16_neon);
         p.cu[BLOCK_32x32].cpy1Dto2D_shr = PFX(cpy1Dto2D_shr_32x32_neon);
 
+        // pixel_sub_ps
+        p.cu[BLOCK_4x4].sub_ps   = PFX(pixel_sub_ps_4x4_neon);
+        p.cu[BLOCK_8x8].sub_ps   = PFX(pixel_sub_ps_8x8_neon);
+        p.cu[BLOCK_16x16].sub_ps = PFX(pixel_sub_ps_16x16_neon);
+        p.cu[BLOCK_32x32].sub_ps = PFX(pixel_sub_ps_32x32_neon);
+        p.cu[BLOCK_64x64].sub_ps = PFX(pixel_sub_ps_64x64_neon);
+
+        // chroma sub_ps
+        p.chroma[X265_CSP_I420].cu[BLOCK_420_4x4].sub_ps   = PFX(pixel_sub_ps_4x4_neon);
+        p.chroma[X265_CSP_I420].cu[BLOCK_420_8x8].sub_ps   = PFX(pixel_sub_ps_8x8_neon);
+        p.chroma[X265_CSP_I420].cu[BLOCK_420_16x16].sub_ps = PFX(pixel_sub_ps_16x16_neon);
+        p.chroma[X265_CSP_I420].cu[BLOCK_420_32x32].sub_ps = PFX(pixel_sub_ps_32x32_neon);
+        p.chroma[X265_CSP_I422].cu[BLOCK_422_4x8].sub_ps   = PFX(pixel_sub_ps_4x8_neon);
+        p.chroma[X265_CSP_I422].cu[BLOCK_422_8x16].sub_ps  = PFX(pixel_sub_ps_8x16_neon);
+        p.chroma[X265_CSP_I422].cu[BLOCK_422_16x32].sub_ps = PFX(pixel_sub_ps_16x32_neon);
+        p.chroma[X265_CSP_I422].cu[BLOCK_422_32x64].sub_ps = PFX(pixel_sub_ps_32x64_neon);
+
         // calc_Residual
         p.cu[BLOCK_4x4].calcresidual[NONALIGNED]   = PFX(getResidual4_neon);
         p.cu[BLOCK_8x8].calcresidual[NONALIGNED]   = PFX(getResidual8_neon);
diff --git a/source/common/aarch64/pixel-util.S b/source/common/aarch64/pixel-util.S
index c7addf2..ea03ed3 100644
--- a/source/common/aarch64/pixel-util.S
+++ b/source/common/aarch64/pixel-util.S
@@ -192,6 +192,176 @@ function x265_getResidual32_neon
     ret
 endfunc
 
+// void pixel_sub_ps_neon(int16_t* a, intptr_t dstride, const pixel* b0, const pixel* b1, intptr_t sstride0, intptr_t sstride1)
+function x265_pixel_sub_ps_4x4_neon
+    lsl             x1, x1, #1
+.rept 2
+    ld1             {v0.8b}, [x2], x4
+    ld1             {v1.8b}, [x3], x5
+    ld1             {v2.8b}, [x2], x4
+    ld1             {v3.8b}, [x3], x5
+    usubl           v4.8h, v0.8b, v1.8b
+    usubl           v5.8h, v2.8b, v3.8b
+    st1             {v4.4h}, [x0], x1
+    st1             {v5.4h}, [x0], x1
+.endr
+    ret
+endfunc
+
+function x265_pixel_sub_ps_8x8_neon
+    lsl             x1, x1, #1
+.rept 4
+    ld1             {v0.8b}, [x2], x4
+    ld1             {v1.8b}, [x3], x5
+    ld1             {v2.8b}, [x2], x4
+    ld1             {v3.8b}, [x3], x5
+    usubl           v4.8h, v0.8b, v1.8b
+    usubl           v5.8h, v2.8b, v3.8b
+    st1             {v4.8h}, [x0], x1
+    st1             {v5.8h}, [x0], x1
+.endr
+    ret
+endfunc
+
+function x265_pixel_sub_ps_16x16_neon
+    lsl             x1, x1, #1
+.rept 8
+    ld1             {v0.16b}, [x2], x4
+    ld1             {v1.16b}, [x3], x5
+    ld1             {v2.16b}, [x2], x4
+    ld1             {v3.16b}, [x3], x5
+    usubl           v4.8h, v0.8b, v1.8b
+    usubl2          v5.8h, v0.16b, v1.16b
+    usubl           v6.8h, v2.8b, v3.8b
+    usubl2          v7.8h, v2.16b, v3.16b
+    st1             {v4.8h-v5.8h}, [x0], x1
+    st1             {v6.8h-v7.8h}, [x0], x1
+.endr
+    ret
+endfunc
+
+function x265_pixel_sub_ps_32x32_neon
+    lsl             x1, x1, #1
+    mov             w12, #4
+.loop_sub_ps_32:
+    sub             w12, w12, #1
+.rept 4
+    ld1             {v0.16b-v1.16b}, [x2], x4
+    ld1             {v2.16b-v3.16b}, [x3], x5
+    ld1             {v4.16b-v5.16b}, [x2], x4
+    ld1             {v6.16b-v7.16b}, [x3], x5
+    usubl           v16.8h, v0.8b, v2.8b
+    usubl2          v17.8h, v0.16b, v2.16b
+    usubl           v18.8h, v1.8b, v3.8b
+    usubl2          v19.8h, v1.16b, v3.16b
+    usubl           v20.8h, v4.8b, v6.8b
+    usubl2          v21.8h, v4.16b, v6.16b
+    usubl           v22.8h, v5.8b, v7.8b
+    usubl2          v23.8h, v5.16b, v7.16b
+    st1             {v16.8h-v19.8h}, [x0], x1
+    st1             {v20.8h-v23.8h}, [x0], x1
+.endr
+    cbnz            w12, .loop_sub_ps_32
+    ret
+endfunc
+
+function x265_pixel_sub_ps_64x64_neon
+    lsl             x1, x1, #1
+    sub             x1, x1, #64
+    mov             w12, #16
+.loop_sub_ps_64:
+    sub             w12, w12, #1
+.rept 4
+    ld1             {v0.16b-v3.16b}, [x2], x4
+    ld1             {v4.16b-v7.16b}, [x3], x5
+    usubl           v16.8h, v0.8b, v4.8b
+    usubl2          v17.8h, v0.16b, v4.16b
+    usubl           v18.8h, v1.8b, v5.8b
+    usubl2          v19.8h, v1.16b, v5.16b
+    usubl           v20.8h, v2.8b, v6.8b
+    usubl2          v21.8h, v2.16b, v6.16b
+    usubl           v22.8h, v3.8b, v7.8b
+    usubl2          v23.8h, v3.16b, v7.16b
+    st1             {v16.8h-v19.8h}, [x0], #64
+    st1             {v20.8h-v23.8h}, [x0], x1
+.endr
+    cbnz            w12, .loop_sub_ps_64
+    ret
+endfunc
+
+// chroma sub_ps
+function x265_pixel_sub_ps_4x8_neon
+    lsl             x1, x1, #1
+.rept 4
+    ld1             {v0.8b}, [x2], x4
+    ld1             {v1.8b}, [x3], x5
+    ld1             {v2.8b}, [x2], x4
+    ld1             {v3.8b}, [x3], x5
+    usubl           v4.8h, v0.8b, v1.8b
+    usubl           v5.8h, v2.8b, v3.8b
+    st1             {v4.4h}, [x0], x1
+    st1             {v5.4h}, [x0], x1
+.endr
+    ret
+endfunc
+
+function x265_pixel_sub_ps_8x16_neon
+    lsl             x1, x1, #1
+.rept 8
+    ld1             {v0.8b}, [x2], x4
+    ld1             {v1.8b}, [x3], x5
+    ld1             {v2.8b}, [x2], x4
+    ld1             {v3.8b}, [x3], x5
+    usubl           v4.8h, v0.8b, v1.8b
+    usubl           v5.8h, v2.8b, v3.8b
+    st1             {v4.8h}, [x0], x1
+    st1             {v5.8h}, [x0], x1
+.endr
+    ret
+endfunc
+
+function x265_pixel_sub_ps_16x32_neon
+    lsl             x1, x1, #1
+.rept 16
+    ld1             {v0.16b}, [x2], x4
+    ld1             {v1.16b}, [x3], x5
+    ld1             {v2.16b}, [x2], x4
+    ld1             {v3.16b}, [x3], x5
+    usubl           v4.8h, v0.8b, v1.8b
+    usubl2          v5.8h, v0.16b, v1.16b
+    usubl           v6.8h, v2.8b, v3.8b
+    usubl2          v7.8h, v2.16b, v3.16b
+    st1             {v4.8h-v5.8h}, [x0], x1
+    st1             {v6.8h-v7.8h}, [x0], x1
+.endr
+    ret
+endfunc
+
+function x265_pixel_sub_ps_32x64_neon
+    lsl             x1, x1, #1
+    mov             w12, #8
+.loop_sub_ps_32x64:
+    sub             w12, w12, #1
+.rept 4
+    ld1             {v0.16b-v1.16b}, [x2], x4
+    ld1             {v2.16b-v3.16b}, [x3], x5
+    ld1             {v4.16b-v5.16b}, [x2], x4
+    ld1             {v6.16b-v7.16b}, [x3], x5
+    usubl           v16.8h, v0.8b, v2.8b
+    usubl2          v17.8h, v0.16b, v2.16b
+    usubl           v18.8h, v1.8b, v3.8b
+    usubl2          v19.8h, v1.16b, v3.16b
+    usubl           v20.8h, v4.8b, v6.8b
+    usubl2          v21.8h, v4.16b, v6.16b
+    usubl           v22.8h, v5.8b, v7.8b
+    usubl2          v23.8h, v5.16b, v7.16b
+    st1             {v16.8h-v19.8h}, [x0], x1
+    st1             {v20.8h-v23.8h}, [x0], x1
+.endr
+    cbnz            w12, .loop_sub_ps_32x64
+    ret
+endfunc
+
 .macro x265_satd_4x8_8x4_end_neon
     add             v0.8h, v4.8h, v6.8h
     add             v1.8h, v5.8h, v7.8h
diff --git a/source/common/aarch64/pixel.h b/source/common/aarch64/pixel.h
index 3d8f0f0..fad959e 100644
--- a/source/common/aarch64/pixel.h
+++ b/source/common/aarch64/pixel.h
@@ -129,6 +129,46 @@ void x265_sad_x4_64x32_neon(const pixel* fenc, const pixel* fref0, const pixel*
 void x265_sad_x4_64x48_neon(const pixel* fenc, const pixel* fref0, const pixel* fref1, const pixel* fref2, const pixel* fref3, intptr_t frefstride, int32_t* res);
 void x265_sad_x4_64x64_neon(const pixel* fenc, const pixel* fref0, const pixel* fref1, const pixel* fref2, const pixel* fref3, intptr_t frefstride, int32_t* res);
 
+sse_t x265_pixel_sse_pp_4x4_neon(const pixel* pix1, intptr_t stride_pix1, const pixel* pix2, intptr_t stride_pix2);
+sse_t x265_pixel_sse_pp_8x8_neon(const pixel* pix1, intptr_t stride_pix1, const pixel* pix2, intptr_t stride_pix2);
+sse_t x265_pixel_sse_pp_16x16_neon(const pixel* pix1, intptr_t stride_pix1, const pixel* pix2, intptr_t stride_pix2);
+sse_t x265_pixel_sse_pp_32x32_neon(const pixel* pix1, intptr_t stride_pix1, const pixel* pix2, intptr_t stride_pix2);
+sse_t x265_pixel_sse_pp_64x64_neon(const pixel* pix1, intptr_t stride_pix1, const pixel* pix2, intptr_t stride_pix2);
+
+sse_t x265_pixel_sse_ss_4x4_neon(const int16_t* pix1, intptr_t stride_pix1, const int16_t* pix2, intptr_t stride_pix2);
+sse_t x265_pixel_sse_ss_8x8_neon(const int16_t* pix1, intptr_t stride_pix1, const int16_t* pix2, intptr_t stride_pix2);
+sse_t x265_pixel_sse_ss_16x16_neon(const int16_t* pix1, intptr_t stride_pix1, const int16_t* pix2, intptr_t stride_pix2);
+sse_t x265_pixel_sse_ss_32x32_neon(const int16_t* pix1, intptr_t stride_pix1, const int16_t* pix2, intptr_t stride_pix2);
+sse_t x265_pixel_sse_ss_64x64_neon(const int16_t* pix1, intptr_t stride_pix1, const int16_t* pix2, intptr_t stride_pix2);
+
+sse_t x265_pixel_ssd_s_4x4_neon(const int16_t* a, intptr_t dstride);
+sse_t x265_pixel_ssd_s_8x8_neon(const int16_t* a, intptr_t dstride);
+sse_t x265_pixel_ssd_s_16x16_neon(const int16_t* a, intptr_t dstride);
+sse_t x265_pixel_ssd_s_32x32_neon(const int16_t* a, intptr_t dstride);
+sse_t x265_pixel_ssd_s_64x64_neon(const int16_t* a, intptr_t dstride);
+
+void x265_pixel_sub_ps_4x4_neon(int16_t* a, intptr_t dstride, const pixel* b0, const pixel* b1, intptr_t sstride0, intptr_t sstride1);
+void x265_pixel_sub_ps_8x8_neon(int16_t* a, intptr_t dstride, const pixel* b0, const pixel* b1, intptr_t sstride0, intptr_t sstride1);
+void x265_pixel_sub_ps_16x16_neon(int16_t* a, intptr_t dstride, const pixel* b0, const pixel* b1, intptr_t sstride0, intptr_t sstride1);
+void x265_pixel_sub_ps_32x32_neon(int16_t* a, intptr_t dstride, const pixel* b0, const pixel* b1, intptr_t sstride0, intptr_t sstride1);
+void x265_pixel_sub_ps_64x64_neon(int16_t* a, intptr_t dstride, const pixel* b0, const pixel* b1, intptr_t sstride0, intptr_t sstride1);
+void x265_pixel_sub_ps_4x8_neon(int16_t* a, intptr_t dstride, const pixel* b0, const pixel* b1, intptr_t sstride0, intptr_t sstride1);
+void x265_pixel_sub_ps_8x16_neon(int16_t* a, intptr_t dstride, const pixel* b0, const pixel* b1, intptr_t sstride0, intptr_t sstride1);
+void x265_pixel_sub_ps_16x32_neon(int16_t* a, intptr_t dstride, const pixel* b0, const pixel* b1, intptr_t sstride0, intptr_t sstride1);
+void x265_pixel_sub_ps_32x64_neon(int16_t* a, intptr_t dstride, const pixel* b0, const pixel* b1, intptr_t sstride0, intptr_t sstride1);
+
+void x265_pixel_add_ps_4x4_neon(pixel* a, intptr_t dstride, const pixel* b0, const int16_t* b1, intptr_t sstride0, intptr_t sstride1);
+void x265_pixel_add_ps_8x8_neon(pixel* a, intptr_t dstride, const pixel* b0, const int16_t* b1, intptr_t sstride0, intptr_t sstride1);
+void x265_pixel_add_ps_16x16_neon(pixel* a, intptr_t dstride, const pixel* b0, const int16_t* b1, intptr_t sstride0, intptr_t sstride1);
+void x265_pixel_add_ps_32x32_neon(pixel* a, intptr_t dstride, const pixel* b0, const int16_t* b1, intptr_t sstride0, intptr_t sstride1);
+void x265_pixel_add_ps_64x64_neon(pixel* a, intptr_t dstride, const pixel* b0, const int16_t* b1, intptr_t sstride0, intptr_t sstride1);
+void x265_pixel_add_ps_4x8_neon(pixel* a, intptr_t dstride, const pixel* b0, const int16_t* b1, intptr_t sstride0, intptr_t sstride1);
+void x265_pixel_add_ps_8x16_neon(pixel* a, intptr_t dstride, const pixel* b0, const int16_t* b1, intptr_t sstride0, intptr_t sstride1);
+void x265_pixel_add_ps_16x32_neon(pixel* a, intptr_t dstride, const pixel* b0, const int16_t* b1, intptr_t sstride0, intptr_t sstride1);
+void x265_pixel_add_ps_32x64_neon(pixel* a, intptr_t dstride, const pixel* b0, const int16_t* b1, intptr_t sstride0, intptr_t sstride1);
+
+void x265_pixel_planecopy_cp_neon(const uint8_t* src, intptr_t srcStride, pixel* dst, intptr_t dstStride, int width, int height, int shift);
+
 void x265_addAvg_4x4_neon(const int16_t* src0, const int16_t* src1, pixel* dst, intptr_t src0Stride, intptr_t src1Stride, intptr_t dstStride);
 void x265_addAvg_4x8_neon(const int16_t* src0, const int16_t* src1, pixel* dst, intptr_t src0Stride, intptr_t src1Stride, intptr_t dstStride);
 void x265_addAvg_4x16_neon(const int16_t* src0, const int16_t* src1, pixel* dst, intptr_t src0Stride, intptr_t src1Stride, intptr_t dstStride);
-- 
2.25.1

